"""Kusto helper functions"""

import pandas
from ._models import KustoResultTable


def dataframe_from_result_table(table):
    """Converts Kusto tables into pandas DataFrame.
    :param azure.kusto.data._models.KustoResultTable table: Table recieved from the response.
    :return: pandas DataFrame.
    :rtype: pandas.DataFrame
    """
    if not table:
        raise ValueError()

    if not isinstance(table, KustoResultTable):
        raise TypeError("Expected KustoResultTable got {}".format(type(table).__name__))

    if not table.columns or not table.rows:
        return pandas.DataFrame()

    frame = pandas.DataFrame.from_records(
        [row.to_list() for row in table.rows], columns=[col.column_name for col in table.columns]
    )
    bool_columns = [col.column_name for col in table.columns if col.column_type == "bool"]
    for col in bool_columns:
        frame[col] = frame[col].astype(bool)

    for i in range(len(table.rows)):
        seventh = table.rows[i]._seventh_digit
        for name in seventh.keys():
            frame.loc[i, name] += pandas.Timedelta(seventh[name] * 100, unit="ns")

    return frame



"""This module constains all classes to get Kusto responses. Including error handling."""

import json

from datetime import timedelta
from abc import ABCMeta, abstractmethod

import six

from ._models import KustoResultColumn, KustoResultRow, KustoResultTable, WellKnownDataSet


@six.add_metaclass(ABCMeta)
class KustoResponseDataSet:
    """Represents the parsed data set carried by the response to a Kusto request."""

    def __init__(self, json_response):
        self.tables = [KustoResultTable(t) for t in json_response]
        self.tables_count = len(self.tables)
        self.tables_names = [t.table_name for t in self.tables]

    @property
    @abstractmethod
    def _error_column(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def _crid_column(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def _status_column(self):
        raise NotImplementedError

    @property
    def primary_results(self):
        """Returns primary results. If there is more than one returns a list."""
        if self.tables_count == 1:
            return self.tables
        primary = list(filter(lambda x: x.table_kind == WellKnownDataSet.PrimaryResult, self.tables))

        return primary

    @property
    def errors_count(self):
        """Checks whether an exception was thrown."""
        query_status_table = next(
            (t for t in self.tables if t.table_kind == WellKnownDataSet.QueryCompletionInformation), None
        )
        if not query_status_table:
            return 0
        min_level = 4
        errors = 0
        for row in query_status_table:
            if row[self._error_column] < 4:
                if row[self._error_column] < min_level:
                    min_level = row[self._error_column]
                    errors = 1
                elif row[self._error_column] == min_level:
                    errors += 1

        return errors

    def get_exceptions(self):
        """Gets the exceptions retrieved from Kusto if exists."""
        query_status_table = next(
            (t for t in self.tables if t.table_kind == WellKnownDataSet.QueryCompletionInformation), None
        )
        if not query_status_table:
            return []
        result = []
        for row in query_status_table:
            if row[self._error_column] < 4:
                result.append(
                    "Please provide the following data to Kusto: CRID='{0}' Description:'{1}'".format(
                        row[self._crid_column], row[self._status_column]
                    )
                )
        return result

    def __iter__(self):
        return iter(self.tables)

    def __getitem__(self, key):
        if isinstance(key, six.integer_types):
            return self.tables[key]
        try:
            return self.tables[self.tables_names.index(key)]
        except ValueError:
            raise LookupError(key)

    def __len__(self):
        return self.tables_count


class KustoResponseDataSetV1(KustoResponseDataSet):

    _status_column = "StatusDescription"
    _crid_column = "ClientActivityId"
    _error_column = "Severity"
    _tables_kinds = {
        "QueryResult": WellKnownDataSet.PrimaryResult,
        "QueryProperties": WellKnownDataSet.QueryProperties,
        "QueryStatus": WellKnownDataSet.QueryCompletionInformation,
    }

    def __init__(self, json_response):
        super(KustoResponseDataSetV1, self).__init__(json_response["Tables"])
        if self.tables_count <= 2:
            self.tables[0].table_kind = WellKnownDataSet.PrimaryResult
            self.tables[0].table_id = 0

            if self.tables_count == 2:
                self.tables[1].table_kind = WellKnownDataSet.QueryProperties
                self.tables[1].table_id = 1
        else:
            toc = self.tables[-1]
            toc.table_kind = WellKnownDataSet.TableOfContents
            toc.table_id = self.tables_count - 1
            for i in range(self.tables_count - 1):
                self.tables[i].table_name = toc[i]["Name"]
                self.tables[i].table_id = toc[i]["Id"]
                self.tables[i].table_kind = self._tables_kinds[toc[i]["Kind"]]


class KustoResponseDataSetV2(KustoResponseDataSet):
    _status_column = "Payload"
    _error_column = "Level"
    _crid_column = "ClientRequestId"

    def __init__(self, json_response):
        super(KustoResponseDataSetV2, self).__init__([t for t in json_response if t["FrameType"] == "DataTable"])



"""Kusto Data Models"""

import json
from datetime import datetime, timedelta
from enum import Enum
from decimal import Decimal
import six
from . import _converters
from .exceptions import KustoServiceError


class WellKnownDataSet(Enum):
    """Categorizes data tables according to the role they play in the data set that a Kusto query returns."""

    PrimaryResult = "PrimaryResult"
    QueryCompletionInformation = "QueryCompletionInformation"
    TableOfContents = "TableOfContents"
    QueryProperties = "QueryProperties"


class KustoResultRow(object):
    """Iterator over a Kusto result row."""

    convertion_funcs = {"datetime": _converters.to_datetime, "timespan": _converters.to_timedelta, "decimal": Decimal}

    def __init__(self, columns, row):
        self._value_by_name = {}
        self._value_by_index = []
        self._seventh_digit = {}
        for i, value in enumerate(row):
            column = columns[i]
            try:
                lower_column_type = column.column_type.lower()
            except AttributeError:
                self._value_by_index.append(value)
                self._value_by_name[columns[i]] = value
                continue

            if lower_column_type in ["datetime", "timespan"]:
                if value is None:
                    typed_value = None
                else:
                    try:
                        # If you are here to read this, you probably hit some datetime/timedelta inconsistencies.
                        # Azure-Data-Explorer(Kusto) supports 7 decimal digits, while the corresponding python types supports only 6.
                        # What we do here, is remove the 7th digit, if exists, and create a datetime/timedelta
                        # from whats left. The reason we are keeping the 7th digit, is to allow users to work with
                        # this precision in case they want it. One example why one might want this precision, is when
                        # working with pandas. In that case, use azure.kusto.data.helpers.dataframe_from_result_table
                        # which takes into account the 7th digit.
                        char = value.split(":")[2].split(".")[1][6]
                        if char.isdigit():
                            tick = int(char)
                            last = value[-1] if value[-1].isalpha() else ""
                            typed_value = KustoResultRow.convertion_funcs[lower_column_type](value[:-2] + last)
                            if tick:
                                if lower_column_type == "datetime":
                                    self._seventh_digit[column.column_name] = tick
                                else:
                                    self._seventh_digit[column.column_name] = (
                                        tick if abs(typed_value) == typed_value else -tick
                                    )
                        else:
                            typed_value = KustoResultRow.convertion_funcs[lower_column_type](value)
                    except (IndexError, AttributeError):
                        typed_value = KustoResultRow.convertion_funcs[lower_column_type](value)
            elif lower_column_type in KustoResultRow.convertion_funcs:
                typed_value = KustoResultRow.convertion_funcs[lower_column_type](value)
            else:
                typed_value = value

            self._value_by_index.append(typed_value)
            self._value_by_name[column.column_name] = typed_value

    @property
    def columns_count(self):
        return len(self._value_by_name)

    def __iter__(self):
        for i in range(self.columns_count):
            yield self[i]

    def __getitem__(self, key):
        if isinstance(key, six.integer_types):
            return self._value_by_index[key]
        return self._value_by_name[key]

    def __len__(self):
        return self.columns_count

    def to_dict(self):
        return self._value_by_name

    def to_list(self):
        return self._value_by_index

    def __str__(self):
        return "['{}']".format("', '".join([str(val) for val in self._value_by_index]))

    def __repr__(self):
        values = [repr(val) for val in self._value_by_name.values()]
        return "KustoResultRow(['{}'], [{}])".format("', '".join(self._value_by_name), ", ".join(values))


class KustoResultColumn(object):
    def __init__(self, json_column, ordianl):
        self.column_name = json_column["ColumnName"]
        self.column_type = json_column.get("ColumnType") or json_column["DataType"]
        self.ordinal = ordianl

    def __repr__(self):
        return "KustoResultColumn({},{})".format(
            json.dumps({"ColumnName": self.column_name, "ColumnType": self.column_type}), self.ordinal
        )


class KustoResultTable(object):
    """Iterator over a Kusto result table."""

    def __init__(self, json_table):
        self.table_name = json_table.get("TableName")
        self.table_id = json_table.get("TableId")
        self.table_kind = WellKnownDataSet[json_table["TableKind"]] if "TableKind" in json_table else None
        self.columns = [KustoResultColumn(column, index) for index, column in enumerate(json_table["Columns"])]

        errors = [row for row in json_table["Rows"] if isinstance(row, dict)]
        if errors:
            raise KustoServiceError(errors[0]["OneApiErrors"][0]["error"]["@message"], json_table)

        self.rows = [KustoResultRow(self.columns, row) for row in json_table["Rows"]]

    @property
    def rows_count(self):
        return len(self.rows)

    @property
    def columns_count(self):
        return len(self.columns)

    def __len__(self):
        return self.rows_count

    def __iter__(self):
        for row in self.rows:
            yield row

    def __getitem__(self, key):
        return self.rows[key]

    def to_dict(self):
        return {"name": self.table_name, "kind": self.table_kind, "data": [r.to_dict() for r in self]}

    def __str__(self):
        d = self.to_dict()
        # enum is not serializable, using value instead
        d["kind"] = d["kind"].value
        return json.dumps(d)
